import argparse
import json
import os
import sys
import logging
import scanResultUtil

parser = argparse.ArgumentParser(
    description="Container image scanning report downloader"
)

parser.add_argument(
    "-t",
    "--target_scan",
    help="JSON scan output of the target image",
    required=True,
    type=str,
)
parser.add_argument(
    "-b",
    "--base_scan",
    help="JSON scan output of the base image",
    required=False,
    type=str,
)
parser.add_argument(
    "-jt",
    "--json_target",
    help="json file of the image being scanned",
    required=False,
    type=str,
)
parser.add_argument(
    "-jb",
    "--json_base",
    help="json file of the baseline image",
    required=False,
    type=str,
)
parser.add_argument(
    "-oh",
    "--output_html",
    help="Output results into an HTML file",
    required=False,
    type=str,
)
parser.add_argument(
    "-oj",
    "--output_json",
    help="Output results into a JSON file",
    required=False,
    type=str,
)
parser.add_argument(
    "-io",
    "--info_os",
    help="Container OS info from /etc/os-release",
    required=False,
    type=str,
)

args = vars(parser.parse_args())

TARGET_SCAN_FILE = None
if args["target_scan"]:
    TARGET_SCAN_FILE = args["target_scan"]

BASE_SCAN_FILE = None
if args["base_scan"]:
    BASE_SCAN_FILE = args["base_scan"]

JSON_TARGET_FILE = None
if args["json_target"]:
    JSON_TARGET_FILE = args["json_target"]

JSON_BASE_FILE = None
if args["json_base"]:
    JSON_BASE_FILE = args["json_base"]

HTML_FILE = None
if args["output_html"]:
    HTML_FILE = args["output_html"]

JSON_FILE = None
if args["output_json"]:
    JSON_FILE = args["output_json"]

logging.basicConfig(
    filename=f"{TARGET_SCAN_FILE[:-5]}.log",
    format="%(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s",
    datefmt="%Y-%m-%d:%H:%M:%S",
    level=logging.DEBUG,
)


def load_file_from_disk_as_json(path):
    try:
        with open(path, "r") as f:
            payload = json.load(f)
            return payload
    except FileNotFoundError as err_file:
        print(err_file)
    except json.decoder.JSONDecodeError as err_json:
        print("Unexpected json format. {}".format(err_json))
    except Exception as err:
        print(err)
    return None


def write_to_file(file_name, data):
    try:
        raw = open(file_name, "w+")
        raw.write(json.dumps(data, indent=4))
    except Exception as err:
        print(err)
        sys.exit(1)
    raw.close()


# Read the target scan json
try:
    scan_report = load_file_from_disk_as_json(TARGET_SCAN_FILE)
    full_report = scan_report.get("vulnerabilities", [])
    packages = scan_report.get("packages", [])
    file_reputation = scan_report.get("file_reputation", [])
    if file_reputation is None:
        # This is an inconsistency with cbctl that we need to manually handle.
        # Check PSEC-4835
        file_reputation = []
except Exception as e:
    full_report = None
    packages = {}
    file_reputation = []

# Get the "sha256:...." to print the link to CBC scan later
image_sha256 = None
if scan_report and "manifest_digest" in scan_report:
    image_sha256 = scan_report["manifest_digest"]

# Read the base scan json
try:
    baseline_report = load_file_from_disk_as_json(BASE_SCAN_FILE)["vulnerabilities"]
except Exception as e:
    baseline_report = []


if full_report is not None:
    if "SAAS_BASE_URL" in os.environ and image_sha256:
        # This can be reused to print the CBC link in the final Summary Table
        os.environ["image_sha256"] = image_sha256
        print(
            f"PSec report available at: https://{os.environ['SAAS_BASE_URL']}/kubernetes/image/{image_sha256}/overview"
        )

    diffed_report = scanResultUtil.report_diff(full_report, baseline_report)

    diffed_report = scanResultUtil.parse_cbctl_binary_vulnerabilities(
        diffed_report, packages
    )

    # Let's normalize the reports
    diffed_report = scanResultUtil.normalize_report(diffed_report)
    baseline_report = scanResultUtil.normalize_report(baseline_report)

    # Only fail/warn on the vulnerabilities in the target image
    scanResultUtil.analyze_vulns(diffed_report)

    # Check if scanned images is based on any of the approved images
    scanResultUtil.check_image_base(JSON_TARGET_FILE, JSON_BASE_FILE)

    # Save the normalized JSON report
    normalized_output = scanResultUtil.normalize_output(diffed_report)
    scanResultUtil.save_json_output(normalized_output, JSON_FILE)

    # Show the os info table
    scanResultUtil.show_info_os_table(HTML_FILE, args.get("info_os", ""))

    # Check file reputation a.k.a. "malware scanning"
    scanResultUtil.check_file_reputation_cbctl(file_reputation, notify=True)
    scanResultUtil.show_malware_table(HTML_FILE)

    scanResultUtil.show_vuln_table(diffed_report, HTML_FILE)
    passed = scanResultUtil.show_report(diffed_report, baseline_report, HTML_FILE)

    if not passed:
        sys.exit(1)
else:
    sys.exit(1)
