#! /usr/bin/env python3

"""PSEC BlackDuck Scanner.

This script uses BlackDuck to perform licensing,
vulnerabilities, and binary analysis against files and docker images.
"""

import argparse
import logging
import os
import re
import requests
import shutil
import subprocess
import sys
import time
import yaml
from git import Repo
from requests.auth import HTTPBasicAuth
from urllib.parse import urlparse
from core import BlackDuckCore

artifactory_user = os.environ.get("ARTIFACTORY_USER")
artifactory_password = os.environ.get("ARTIFACTORY_PASSWORD")
rep_path_cli = os.environ.get("REP_PATH_CLI", ".")
bcom_bduck_url = os.environ.get("BCOM_BDUCK_URL")
bcom_bduck_api_token = os.environ.get("BCOM_BDUCK_API_TOKEN")
github_token = os.environ.get("GITHUB_TOKEN")
sm_repo_url = f"https://{github_token}@github.gwd.broadcom.net/SED/ESG-ProdSec-Inventory.git"
container_projects = ["ESG-CBC-Platform"]

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(levelname)s - [%(filename)s:%(lineno)d] - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    handlers=[
        logging.FileHandler("blackduck.log"),
        logging.StreamHandler(),
    ],
)

logger = logging.getLogger(__name__)


def validate_variables():
    """Validates the presence of required environment variables.

    Raises:
        ValueError: if any required environment variable is not set.
    """

    required_vars = ["BCOM_BDUCK_URL", "BCOM_BDUCK_API_TOKEN"]

    logger.info("Validating required variables ..")

    for var in required_vars:
        if not os.environ.get(var):
            logger.error(f"Required environment variable {var} is not set.")
            sys.exit(1)

    logger.info("All required environment variables are set.")


def clone_repo(repo_url: str, local_dir: str) -> str:
    """Clones a Git repository and returns the path to the service_mapping.yaml file.

    Args:
        repo_url (str): The URL of the Git repository to clone.
        local_dir (str): The local directory where the repository will be cloned.

    Returns:
        str: The path to the service_mapping.yaml file if it exists, otherwise None.
    """

    max_retries = 3
    wait_time = 5

    if os.path.exists(local_dir):
        try:
            shutil.rmtree(local_dir)
            logger.info(f"Removed existing directory before cloning git repo: {local_dir}")
        except Exception as e:
            logger.error(f"Failed to remove local git directory with error: {e}")
            return None

    try:
        Repo.clone_from(repo_url, local_dir)
        logger.info(f"Git repository '{local_dir}' successfully cloned.")
    except Exception as e:
        logger.error(f"Failed to clone repository '{local_dir}' with error: {e}")
        return None

    service_mapping_path = os.path.join(local_dir, "service_mapping.yaml")

    for attempt in range(max_retries):
        if os.path.isfile(service_mapping_path) and os.path.getsize(service_mapping_path) > 0:
            logger.info(f"Found service_mapping.yaml at: {service_mapping_path}")
            return service_mapping_path
        else:
            logger.warning(f"service_mapping.yaml not found or is empty. Attempt {attempt + 1} of {max_retries}")
            time.sleep(wait_time)

    logger.error("service_mapping.yaml not found after multiple attempts.")
    return None


def download_file(url: str) -> str:
    """Downloads target file to be scanned.

    Args:
        url (str): URL of the file to be downloaded.

    Raises:
        RequestException: if the request fails.

    Returns:
        file_path (str): path to the file to be scanned.
    """

    session = requests.Session()
    auth = None

    try:
        if "usw1.packages.broadcom.com" in url:
            auth = HTTPBasicAuth(artifactory_user, artifactory_password)

        file_path = os.path.join(os.getcwd(), os.path.basename(urlparse(url).path))

        with session.get(url, auth=auth, stream=True) as s_response:
            with open(file_path, "wb") as file:
                for chunk in s_response.iter_content(chunk_size=8192):
                    file.write(chunk)

        logger.info(f"File downloaded and saved at: {file_path}")
        return file_path
    except requests.exceptions.RequestException as e:
        logger.error(f"Failed to download the target file: {e}")
        sys.exit(1)


def read_yaml(search_item: str, mapping_file: str, use_case: int) -> str:
    """Traverses the YAML file based on the use case and returns results.

    Args:
        search_item (str): The item to search for (key or container image).
        mapping_file (str): Path to the YAML file to search against.
        use_case (int): The use case to execute (1 for project group search, 2 for container image search).

    Returns:
        str: The found key or Engineering Group value, or None if not found.
    """
    try:
        with open(mapping_file, "r") as f:
            data = yaml.safe_load(f)
    except Exception as e:
        logger.error(f"Failed to load {mapping_file}. Exiting with error: {e}")
        sys.exit(1)

    def traverse(data):
        if isinstance(data, dict):
            for key, value in data.items():
                if use_case == 1:
                    if isinstance(value, list):
                        if search_item in value:
                            return key
                    elif isinstance(value, dict):
                        result = traverse(value)

                        if result:
                            return result
                elif use_case == 2:
                    if key == "Engineering Group":
                        container_images = data.get("Assets", {}).get("Container Images", [])

                        if any(search_item in container_image for container_image in container_images):
                            return value

                if isinstance(value, list):
                    for item in value:
                        if isinstance(item, dict):
                            result = traverse(item)

                            if result:
                                return result
                elif isinstance(value, dict):
                    result = traverse(value)
                    if result:
                        return result
        return None

    return traverse(data)


def get_team_name(image_name: str) -> str:
    """Retrieves the engineering team name associated with a specified container image.

    Args:
        image_name (str): The name of the container image to search for in the YAML file.

    Returns:
        str: The name of the engineering team associated with the specified container image.
             Returns None if the image name is not found in the mapping file.
    """

    logger.info(f"Retrieving team name based on image name: '{image_name}'")

    local_dir = "ESG-ProdSec-Inventory"
    mapping_file = clone_repo(sm_repo_url, local_dir)

    if mapping_file:
        team_name = read_yaml(image_name, mapping_file, 2)

        if team_name:
            logger.info(f"Team name successfully retrieved: {team_name}")

            return team_name
        else:
            logger.error("Failed to retrieve team name. Setting it to 'Undefined'.")

            return "Undefined"
    else:
        logger.error("Failed to clone git repository.")
        sys.exit(1)


def get_project_group(project_name: str) -> str:
    """Determines the project group name.

    Searches for project_name in the mapping file.
    If no matches are found, project_name is set to a default value.

    Args:
        project_name (str): the name of the project to search for.

    Returns:
        str: the project group name.
    """

    mapping_file = os.path.join(os.getcwd(), "blackduck/projects_mapping.yml")
    project_group = read_yaml(project_name, mapping_file, 1)

    if project_group is None:
        project_group = "CB Cloud (CBC)"

    return project_group


def run_synopsys_detect(
    project_name: str,
    project_version: str,
    project_group: str,
    detect_tools: str,
    excluded_tools: str,
    logging_level: str,
    target: str,
):
    """Runs BlackDuck Detect via shell command.

    Invokes BlackDuck Detect and scans a given target
    using several different detectors.

    Args:
        project_name (str): name of the project in BlackDuck HUB.
        project_version (str): version of the project in BlackDuck HUB.
        project_group (str): name of the project group in BlackDuck HUB.
        detect_tools (str): Detect tools to include in the scan.
        excluded_tools (str): Detect tools to exclude from the scan.
        target (str): file, directory or container image to be scanned.

    Raises:
        CalledProcessError: If the subprocess command fails.

    Returns:
        file_path (str): Path to the file to be scanned.
    """

    project_group = project_group or get_project_group(project_name)
    project_name = project_name or "unknown"
    project_version = project_version or "unknown"
    target = target
    bduck_detect_tools = detect_tools or "BINARY_SCAN"
    bduck_excluded_tools = (
        excluded_tools
        or "DETECTOR,SIGNATURE_SCAN,IMPACT_ANALYSIS,DOCKER,BAZEL,IAC_SCAN,CONTAINER_SCAN,THREAT_INTEL"
    )
    bduck_clone_categories = "COMPONENT_DATA,CUSTOM_FIELD_DATA,DEEP_LICENSE,LICENSE_TERM_FULFILLMENT,VULN_DATA"
    bduck_logging_level = logging_level or "INFO"
    blackduck_core = BlackDuckCore()

    image_name_pattern = re.compile(
        r"^([a-z0-9.-]+(:[a-zA-Z0-9.-]+)?(/[a-zA-Z0-9.-]+)+(:[a-zA-Z0-9.-]+)?)$"
    )

    if not blackduck_core.project_exists(project_name):
        logger.error(f"The project '{project_name}' does not exist in BlackDuck. Exiting ..")
        sys.exit(1)

    cmd_args = {
        "--blackduck.url": bcom_bduck_url,
        "--blackduck.api.token": bcom_bduck_api_token,
        "--blackduck.trust.cert": "true",
        "--detect.project.group.name": project_group,
        "--detect.project.name": project_name,
        "--detect.project.version.name": project_version,
        "--detect.tools": bduck_detect_tools,
        "--detect.tools.excluded": bduck_excluded_tools,
        "--logging.level.com.synopsys.integration": bduck_logging_level,
        "--detect.risk.report.pdf": "false",
        "--detect.wait.for.results": "false",
        "--detect.timeout": "1200",
        "--detect.cleanup": "true",
        "--detect.clone.project.version.latest": "true",
        "--detect.project.clone.categories": bduck_clone_categories,
        "--detect.project.version.update": "true",
    }

    if os.path.isdir(target):
        logger.error(f"Provide the full path to the binary file. A path to a directory was provided instead.")
        sys.exit(1)
    elif os.path.isfile(target):
        destination_path = os.path.join(rep_path_cli, os.path.basename(target))

        if not os.path.exists(destination_path):
            try:
                target = shutil.copy2(target, f"{rep_path_cli}/")
                cmd_args["--detect.binary.scan.file.path"] = target
            except Exception as e:
                logger.error(f"Failed to copy the target file: {e}")
                sys.exit(1)
        else:
            logger.info("Target file already exists in the destination. Proceeding ..")
            cmd_args["--detect.binary.scan.file.path"] = destination_path
    elif target.startswith(("http://", "https://")):
        target = download_file(target)

        try:
            target = shutil.copy2(target, f"{rep_path_cli}/")
            cmd_args["--detect.binary.scan.file.path"] = target
        except Exception as e:
            logger.error(f"Failed to copy the target file: {e}")
            sys.exit(1)
    elif image_name_pattern.match(target):
        if project_name in container_projects:
            match = re.search(r'/(?:[^/]+/)(.+?):', target)
            image_basename = match.group(1)
            team_name = get_team_name(image_basename)
            project_version = f"ESG-CBC-Team-{team_name}"

        if project_version.lower() == "unknown" or project_version == "" or project_version is None:
            project_version = target.rsplit(":", 1)[0]

        cmd_args["--detect.project.version.name"] = project_version

        file_name = target.rsplit(":", 1)[0]
        file_path = os.path.join(os.getcwd(), os.path.basename(file_name))
        cmd_args["--detect.project.version.distribution"] = "SAAS"

        if "container_scan" in bduck_detect_tools.lower():
            cmd_args["--detect.container.scan.file.path"] = f"{file_path}"
        if "docker" in bduck_detect_tools.lower():
            cmd_args["--detect.docker.tar"] = f"{file_path}"

        try:
            logger.info(f"Pulling {target}")
            subprocess.run(["docker", "pull", f"{target}"], check=True)

            logger.info(f"Saving {target} as a tar file")
            subprocess.run(["docker", "save", "--output", f"{file_path}", f"{target}"], check=True)
        except Exception as e:
            logger.error(f"Command execution failed. Error: {e}")
            sys.exit(1)
    else:
        logger.error("Provide a file URL, local file path, or container image name.")
        sys.exit(1)

    cmd = ["java", "-jar", "/root/synopsys-detect.jar"]

    for arg, value in cmd_args.items():
        cmd.append(f"{arg}={value}")

    try:
        subprocess.run(cmd, check=True)
    except Exception as e:
        logger.error(f"Command execution failed. Error: {e}. Command: {' '.join(cmd)}")
        sys.exit(1)

    try:
        project_id = blackduck_core.get_project_id(project_name)
        versions = blackduck_core.get_project_versions(project_id, project_name)

        if not versions:
            logger.info(f"The project '{project_name}' does not contain any versions. Skipping version deprecation and update.")
            sys.exit(0)

        versions.sort(key=lambda v: v["createdAt"])

        if project_name not in container_projects:
            released_versions = [v for v in versions if v["phase"].lower() == "released"]

            if released_versions:
                released_versions.sort(key=lambda v: v["createdAt"])

                for rv in released_versions:
                    released_version_url = rv["_meta"]["href"]
                    released_version_name = rv["versionName"]

                    blackduck_core.update_version_phase(
                        project_id,
                        released_version_url,
                        released_version_name,
                        "DEPRECATED",
                    )
            else:
                logger.info(f"No versions in RELEASED phase found for '{project_name}'! Skipping version deprecation.")

            latest_version_url = versions[-1]["_meta"]["href"]
            latest_version_name = versions[-1]["versionName"]

            if versions[-1]["phase"].lower() != "released":
                blackduck_core.update_version_phase(project_id, latest_version_url, latest_version_name, "RELEASED")
            else:
                logger.info(f"The latest version '{latest_version_name}' is already marked as RELEASED.")
        else:
            non_released_versions = [v for v in versions if v["phase"].lower() != "released"]

            if non_released_versions:
                for nrv in non_released_versions:
                    nrv_url = nrv["_meta"]["href"]
                    nrv_name = nrv["versionName"]

                    blackduck_core.update_version_phase(project_id, nrv_url, nrv_name, "RELEASED")
            else:
                logger.info(f"All versions for '{project_name}' are already marked as RELEASED.")
    except Exception as e:
        logger.error(f"Failed to retrieve project versions or update phases: {e}")
        sys.exit(1)


def main():
    validate_variables()

    parser = argparse.ArgumentParser(description="PSEC BlackDuck Scanner")

    parser.add_argument(
        "-pn",
        "--project_name",
        help="Project name in BlackDuck",
        required=False,
        type=str,
    )
    parser.add_argument(
        "-pv",
        "--project_version",
        help="Project version in BlackDuck",
        required=False,
        type=str,
    )
    parser.add_argument(
        "-pg",
        "--project_group",
        help="Project parent group name in BlackDuck",
        required=False,
        type=str,
    )
    parser.add_argument(
        "-dt",
        "--detect_tools",
        help="Detect tools to include in the scan",
        required=False,
        type=str,
    )
    parser.add_argument(
        "-et",
        "--excluded_tools",
        help="Detect tools to exclude from the scan",
        required=False,
        type=str,
    )
    parser.add_argument(
        "-t",
        "--target",
        help="File URL, local file path, or full container image URL",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-ll",
        "--logging_level",
        help="BlackDuck logging level",
        required=False,
        type=str,
    )

    args = parser.parse_args()
    project_name = args.project_name
    project_version = args.project_version
    project_group = args.project_group
    detect_tools = args.detect_tools
    excluded_tools = args.excluded_tools
    logging_level = args.logging_level
    target = args.target

    logger.info(f"Initiating BlackDuck scan activities ..")

    for t in target.split():
        run_synopsys_detect(
            project_name,
            project_version,
            project_group,
            detect_tools,
            excluded_tools,
            logging_level,
            t,
        )

    logger.info("BlackDuck scan activities are now complete.")


if __name__ == "__main__":
    main()
